import json
import logging
from contextlib import contextmanager

import requests
import requests.exceptions
from tenacity import retry, wait_fixed, stop_after_attempt


logger = logging.getLogger(__name__)


class Imposter(object):
    """Represents a mountebank imposter.

    This is used to mock services.

    Usage:
        imposter = Imposter()
        imposter.add_stub("/test", "GET", "What I'm expecting")
        with imposter.mockhttp() as url:
            assert requests.get(url + "/test").text == "What I'm expecting"
    """

    def __init__(self, mountebank_scheme="http", mountebank_host="mountebank",
        mountebank_port=2525):
        """Instantiates a new Imposter instance.

        Args:
            mountebank_scheme (str): Optional scheme to use to connect with
            mountebank.
            mountebank_host (str): Optional host to use to connect with
            mountebank.
            mountebank_port (int): Optional port to use to connect with
            mountebank.
        """
        self.port = None
        self.mountebank_scheme = mountebank_scheme
        self.mountebank_host = mountebank_host
        self.mountebank_port = mountebank_port
        self.mountebank_url = "%s://%s:%s" % (
            mountebank_scheme, mountebank_host, mountebank_port)

        self._imposter = {
            "protocol": "http",
            "stubs": [{
                "responses": [
                    {"is": {"statusCode": 404}}
                ]
            }]
        }

    def add_stub(self, path, method, body, repeat=0, status_code=200):
        """Adds a stub to mock.

        Args:
            path (str): The path being mocked.
            path (str): The method being mocked.
            body (str): The text the mock should return.
            repeat (int): The total number of times to repeat a response. If
            not set then the response will repeat indefinitely.
            status_code (int): The status code to return.
        """
        predicates = [{
            "equals": {
                "method": method,
                "path": path
            }
        }]
        responses = [{
            "is": {
                "statusCode": status_code,
                "body": body
            }
        }]

        if repeat > 0:
            responses = responses * repeat

        # Check to see if a stub already exists for the predicates. If so,
        # update the existing stub.
        predicates_exists = False
        for stub in self._imposter["stubs"]:
            if "predicates" in stub and stub["predicates"] == predicates:
                predicates_exists = True
                stub["responses"] += responses
                break

        # Add the new stub near the bottom before our catch-all
        if not predicates_exists:
            self._imposter["stubs"].insert(len(self._imposter["stubs"]) - 1, {
                "predicates": predicates,
                "responses": responses
            })

    @contextmanager
    def mockhttp(self, port=None, wait_for_mountebank=True):
        """A contextmanager that uses mountebank to mock out a service.

        Args:
            port (int): Optional static port to open for the new mock service.
            If not provided a port will be randomly generated.
            wait_for_mountebank (bool): Whether or not to wait for mountebank
            to become available. This is useful in containers where mountebank
            might still be loading while tests are running. By default this
            function will wait a few seconds for mountebank to become
            available. Set to False to not wait at all.

        Yields:
            str: The URL to the mocked service.
        """
        if wait_for_mountebank:
            self.wait()

        if port:
            self._imposter["port"] = port

        response = requests.post(self.mountebank_url + "/imposters",
            data=json.dumps(self._imposter))

        try:
            response.raise_for_status()

        except requests.exceptions.HTTPError:
            logger.exception("Mountebank returned: %s", response.text)
            raise

        self.port = response.json()["port"]
        url = "%s://%s:%s" % (
            self.mountebank_scheme, self.mountebank_host, self.port)

        try:
            yield url

        finally:
            response = requests.delete("%s/imposters/%s" % (
                self.mountebank_url, self.port))

            try:
                response.raise_for_status()

            except requests.exceptions.HTTPError:
                logger.exception(
                    "Unable to delete mountebank mock: %s", response.text)
                raise

    def get_requests(self):
        """Retrieves the requests received by the mock service.

        Returns:
            dict: The requests received by the mock service.
        """
        response = requests.get("%s/imposters/%s" % (
            self.mountebank_url, self.port))
        response_data = response.json()
        return response_data["requests"]

    def wait(self):
        """Blocks until mountebank is accessible (or waiting times out).
        """
        def wait():
            requests.get(self.mountebank_url).raise_for_status()

        retry(wait=wait_fixed(1), stop=stop_after_attempt(10))(wait)()



from setuptools import setup, find_packages


with open("README.rst") as readme_file:
    readme = readme_file.read()


with open("requirements.txt") as requirements_file:
    requirements = [line.strip() for line in requirements_file]


setup(
    name="pymountebank",
    description="Mountebank wrapper for mocking HTTP APIs",
    license="BSD 3-Clause License",
    version="1.4.1",
    author="Quid Inc",
    author_email="infrastructure@quid.com",
    url="https://github.com/quid/pymountebank",
    packages=find_packages(exclude=["tests*"]),
    include_package_data=True,
    long_description=readme,
    install_requires=requirements,
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ]
)



import requests

from mountebank import Imposter


def test_example():
    imposter = Imposter()
    imposter.add_stub("/test", "GET", "What I'm expecting")
    with imposter.mockhttp() as url:
        response = requests.get(url + "/test")
        response.raise_for_status()
        assert response.text == "What I'm expecting"

        response = requests.get(url + "/test")
        response.raise_for_status()
        assert response.text == "What I'm expecting"


def test_status_codes():

    imposter = Imposter()
    imposter.add_stub("/test", "GET", "Normal response")
    imposter.add_stub(
        "/test",
        "GET",
        "Response with a status code",
        status_code=418
    )

    with imposter.mockhttp() as url:
        response = requests.get(url + "/test")
        assert response.text == "Normal response"
        assert response.status_code == 200

        response = requests.get(url + "/test")
        assert response.text == "Response with a status code"
        assert response.status_code == 418


def test_repeat():
    imposter = Imposter()
    imposter.add_stub("/test", "GET", "Not what I'm expecting", repeat=2)
    imposter.add_stub("/test", "GET", "What I'm expecting")

    with imposter.mockhttp() as url:
        response = requests.get(url + "/test")
        assert response.text == "Not what I'm expecting"
        response = requests.get(url + "/test")
        assert response.text == "Not what I'm expecting"
        response = requests.get(url + "/test")
        assert response.text == "What I'm expecting"

        # Requests loop around when there's repetition
        response = requests.get(url + "/test")
        assert response.text == "Not what I'm expecting"


def test_implied_repeat():
    imposter = Imposter()
    imposter.add_stub("/test", "GET", "Not what I'm expecting")
    imposter.add_stub("/test", "GET", "What I'm expecting")

    with imposter.mockhttp() as url:
        response = requests.get(url + "/test")
        assert response.text == "Not what I'm expecting"

        response = requests.get(url + "/test")
        assert response.text == "What I'm expecting"

        response = requests.get(url + "/test")
        assert response.text == "Not what I'm expecting"


def test_get_requests():
    imposter = Imposter()
    with imposter.mockhttp() as url:
        requests.get(url + "/test")
        requests.get(url + "/anothertest")
        requests.get(url + "/why/not/another")

        mocked_requests = imposter.get_requests()
        assert len(mocked_requests) == 3
        assert mocked_requests[0]["path"] == "/test"
        assert mocked_requests[1]["path"] == "/anothertest"
        assert mocked_requests[2]["path"] == "/why/not/another"


def test_get_no_requests():
    imposter = Imposter()
    with imposter.mockhttp() as url:
        assert imposter.get_requests() == []


def test_static_port():
    imposter = Imposter()
    imposter.add_stub("/test", "GET", "What I'm expecting")
    with imposter.mockhttp(8080) as url:
        assert url == "http://mountebank:8080"
        assert imposter.port == 8080
        response = requests.get(url + "/test")
        response.raise_for_status()
        assert response.text == "What I'm expecting"
